/*	PANDA -- a simple transaction monitor

Copyright (C) 2001-2003 Ogochan & JMA (Japan Medical Association).

This module is part of PANDA.

	PANDA is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
to anyone for the consequences of using it or for whether it serves
any particular purpose or works at all, unless he says so in writing.
Refer to the GNU General Public License for full details. 

	Everyone is granted permission to copy, modify and redistribute
PANDA, but only under the conditions described in the GNU General
Public License.  A copy of this license is supposed to have been given
to you along with PANDA so you can know your rights and
responsibilities.  It should be in a file named COPYING.  Among other
things, the copyright notice and this notice must be preserved on all
copies. 
*/

#define	MAIN
/*
#define	DEBUG
#define	TRACE
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
#include	<signal.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<string.h>
#include	<libgen.h>
#include    <sys/socket.h>
#include	<pthread.h>
#include	<glib.h>

#include	"types.h"
#include	"const.h"
#include	"libmondai.h"
#include	"enum.h"
#include	"dirs.h"
#include	"socket.h"
#include	"net.h"
#include	"comm.h"
#include	"directory.h"
#include	"dblib.h"
#include	"dbgroup.h"
#include	"handler.h"
#include	"wfc.h"
#include	"apsio.h"
#include	"aps_main.h"
#include	"option.h"
#include	"message.h"
#include	"debug.h"

static	Bool	fConnect;
static	sigset_t	hupset;
static	int		MaxTran;
static	int		Sleep;

static	void
InitSystem(
	char	*name)
{
	int		i;

dbgmsg(">InitSystem");
	sigemptyset(&hupset); 
	sigaddset(&hupset,SIGHUP);
	pthread_sigmask(SIG_BLOCK,&hupset,NULL);
	InitDirectory();
dbgmsg("*");
	SetUpDirectory(Directory,name,"","");
dbgmsg("*");
	if		(  ( ThisLD = GetLD(name) )  ==  NULL  ) {
		dbgprintf("LD \"%s\" not found.",name);
		exit(1);
	}
	if		(  ThisLD->home  !=  NULL  ) {
		chdir(ThisLD->home);
	}
	ThisBD = NULL;
	ThisDBD = NULL;

dbgmsg("*");
	InitiateHandler();
dbgmsg("*");
	ThisDB = ThisLD->db;
	DB_Table = ThisLD->DB_Table;
	TextSize = ThisLD->textsize;
	for	( i = 0 ; i < ThisLD->cWindow ; i ++ ) {
		dbgprintf("[%s]",ThisLD->window[i]->rec->name);
		InitializeValue(ThisLD->window[i]->rec->value);
	}
dbgmsg("*");
	ReadyDC();
dbgmsg("*");
	if		(  ThisLD->cDB  >  0  ) {
		ReadyOnlineDB();
	}
dbgmsg("<InitSystem");
}

static	ProcessNode	*
MakeProcessNode(void)
{
	ProcessNode	*node;
	int			i;

dbgmsg(">MakeProcessNode");
	node = New(ProcessNode);
	node->mcprec = ThisEnv->mcprec;
	node->linkrec = ThisEnv->linkrec;
	node->sparec = ThisLD->sparec;
	node->cWindow = ThisLD->cWindow;
	node->whash = ThisLD->whash;
	node->textsize = ThisLD->textsize;
	node->scrrec = (RecordStruct **)xmalloc(sizeof(RecordStruct *) * node->cWindow);
	for	( i = 0 ; i < node->cWindow ; i ++ ) {
		node->scrrec[i] = ThisLD->window[i]->rec;
	}

	/*	get initialize memory area	*/

	SetValueInteger(GetItemLongName(node->mcprec->value,"private.pstatus"),APL_SESSION_LINK);
	SetValueInteger(GetItemLongName(node->mcprec->value,"private.pputtype"),0);
	SetValueInteger(GetItemLongName(node->mcprec->value,"private.prc"),0);
dbgmsg("<MakeProcessNode");
	return	(node);
}

static	void
FinishSession(
	ProcessNode	*node)
{
dbgmsg(">FinishSession");
	xfree(node->scrrec); 
	xfree(node);
dbgmsg("<FinishSession");
}

static	void
ExecuteDC(
	void	*para)
{
	int		fhWFC
	,		_fh;
	NETFILE	*fpWFC;

	ProcessNode	*node;
	WindowBind	*bind;
	int		tran;

ENTER_FUNC;
	if		(  !fConnect  )	{
		_fh = InitServerPort(PortNumber,Back);
		if		(  ( fhWFC = accept(_fh,0,0) )  <  0  )	{
			Error("INET Domain Accept");
		}
		fpWFC = SocketToNet(fhWFC);
	} else {
		if		(  WfcPortNumber  ==  NULL  ) {
			if		(  ThisLD->wfc  !=  NULL  ) {
				WfcPortNumber = ThisLD->wfc->port;
			} else {
				WfcPortNumber = ThisEnv->WfcApsPort->port;
			}
		}
		if		(  WFC_Host  ==  NULL  ) {
			if		(  ThisLD->wfc  !=  NULL  ) {
				WFC_Host = ThisLD->wfc->host;
			} else {
				WFC_Host = ThisEnv->WfcApsPort->host;
			}
		}
		dbgprintf("%s:%s",WFC_Host,WfcPortNumber);
		if		(  ( fhWFC = ConnectSocket(WfcPortNumber,SOCK_STREAM,WFC_Host) )
				   <  0  ) {
			Error("WFC not ready");
		}
		fpWFC = SocketToNet(fhWFC);
		SendString(fpWFC,ThisLD->name);
		if		(  RecvPacketClass(fpWFC)  !=  APS_OK  ) {
			Error("invalid LD name");
		}
	}
	InitAPSIO(fpWFC);
	node = MakeProcessNode();
	for	( tran = MaxTran;(	(  MaxTran  ==  0  )
						||	(  tran     >   0  ) ); tran -- ) {
		if		(  !GetWFC(fpWFC,node)  )	break;
		dbgprintf("[%s]",ThisLD->name);
		if		(  ( bind = (WindowBind *)g_hash_table_lookup(ThisLD->whash,
															  ValueStringPointer(GetItemLongName(node->mcprec->value,"dc.window"))))  !=  NULL  ) {
			if		(  bind->module  ==  NULL  )	break;
			SetValueString(GetItemLongName(node->mcprec->value,"dc.module"),bind->module,NULL);
			ExecDBG_Operation(NULL,"DBSTART");
			ExecuteProcess(node);
			if		(  Sleep  >  0  ) {
				sleep(Sleep);
			}
			ExecDBG_Operation(NULL,"DBCOMMIT");
			PutWFC(fpWFC,node);
		} else {
			MessagePrintf("window [%s] not found.\n",
						  ValueToString(GetItemLongName(node->mcprec->value,"dc.window"),NULL));
			break;
		}
	}
	MessageLog("exiting DC_Thread\n");
	CloseNet(fpWFC);
	FinishSession(node);
LEAVE_FUNC;
}

static	void
ExecuteServer(void)
{
dbgmsg(">ExecuteServer");
	ExecuteDC(NULL);
dbgmsg("<ExecuteServer");
}

static	void
StopProcess(
	int		ec)
{
dbgmsg(">StopProcess");
	StopOnlineDB(); 
	CleanUpOnlineDB(); 
	StopDC();
	CleanUpOnlineDC();
dbgmsg("<StopProcess");
	exit(ec);
}

static	ARG_TABLE	option[] = {
	{	"port",		STRING,		TRUE,	(void*)&PortNumber,
		"ݡֹ"	 								},
	{	"back",		INTEGER,	TRUE,	(void*)&Back,
		"³Ԥ塼ο" 							},

	{	"connect",	BOOLEAN,	TRUE,	(void*)&fConnect,
		"ͥȥ⡼(apswfc³)"			},
	{	"host",		STRING,		TRUE,	(void*)&WFC_Host,
		"WFCƯۥ̾"								},
	{	"wfcport",	STRING,		TRUE,	(void*)&WfcPortNumber,
		"WFC³Ԥݡֹ"	 						},

	{	"base",		STRING,		TRUE,	(void*)&BaseDir,
		"ĶΥ١ǥ쥯ȥ"		 				},
	{	"record",	STRING,		TRUE,	(void*)&RecordDir,
		"ǡǼǥ쥯ȥ"	 				},
	{	"lddir",	STRING,		TRUE,	(void*)&D_Dir,
		"LDǼǥ쥯ȥ"		 				},
	{	"dir",		STRING,		TRUE,	(void*)&Directory,
		"ǥ쥯ȥե"	 						},
	{	"path",		STRING,		TRUE,	(void*)&LibPath,
		"⥸塼Υɥѥ"						},

	{	"dbhost",	STRING,		TRUE,	(void*)&DB_Host,
		"ǡ١Ưۥ̾"						},
	{	"dbport",	STRING,		TRUE,	(void*)&DB_Port,
		"ǡ١Եݡֹ"					},
	{	"db",		STRING,		TRUE,	(void*)&DB_Name,
		"ǡ١̾"								},
	{	"dbuser",	STRING,		TRUE,	(void*)&DB_User,
		"ǡ١Υ桼̾"						},
	{	"dbpass",	STRING,		TRUE,	(void*)&DB_Pass,
		"ǡ١Υѥ"						},

	{	"maxtran",	INTEGER,	TRUE,	(void*)&MaxTran,
		"apsνȥ󥶥ꤹ"		},
	{	"cache",	INTEGER,	TRUE,	(void*)&nCache,
		"ü򥭥å夹ü"				},

	{	"sleep",	INTEGER,	TRUE,	(void*)&Sleep,
		"¹Ի֤­(for debug)"				},

	{	NULL,		0,			FALSE,	NULL,	NULL 	}
};

static	void
SetDefault(void)
{
	WfcPortNumber = NULL;
	WFC_Host = NULL;
	PortNumber = IntStrDup(PORT_APS_BASE);
	Back = 5;
	fConnect = FALSE;

	BaseDir = NULL;
	RecordDir = NULL;
	D_Dir = NULL;
	Directory = "./directory";
	LibPath = NULL;
	MaxTran = 0;
	nCache = 0;
	Sleep = 0;

	DB_User = NULL;
	DB_Pass = NULL;
	DB_Host = NULL;
	DB_Port = NULL;
	DB_Name = DB_User;
}

extern	int
main(
	int		argc,
	char	**argv)
{
	FILE_LIST	*fl;
	int			rc;
	char		id[128];

	SetDefault();
	fl = GetOption(option,argc,argv);
	(void)signal(SIGHUP,(void *)StopProcess);
	if		(	(  fl  !=  NULL  )
			&&	(  fl->name  !=  NULL  ) ) {
		sprintf(id,"aps-%s",fl->name);
		InitMessage(id,NULL);
		InitNET();
		InitSystem(fl->name);
dbgmsg("-\n---\n");
		ExecuteServer();
		StopProcess(0);
		rc = 0;
	} else {
		rc = -1;
		fprintf(stderr,"LD̾ꤵƤޤ\n");
	}
	exit(rc);
}
