/*	PANDA -- a simple transaction monitor

Copyright (C) 1998-1999 Ogochan.
              2000-2003 Ogochan & JMA (Japan Medical Association).

This module is part of PANDA.

	PANDA is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
to anyone for the consequences of using it or for whether it serves
any particular purpose or works at all, unless he says so in writing.
Refer to the GNU General Public License for full details. 

	Everyone is granted permission to copy, modify and redistribute
PANDA, but only under the conditions described in the GNU General
Public License.  A copy of this license is supposed to have been given
to you along with PANDA so you can know your rights and
responsibilities.  It should be in a file named COPYING.  Among other
things, the copyright notice and this notice must be preserved on all
copies. 
*/

/*
#define	DEBUG
#define	TRACE
#define	MAIN
*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#define	_STYLE_PARSER

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>
#include	<unistd.h>
#include	<glib.h>
#include	<sys/stat.h>
#include	<gtk/gtk.h>
#include	"types.h"
#include	"libmondai.h"
#include	"styleLex.h"
#include	"styleParser.h"
#include	"debug.h"

static	char	*StyleFileName;
static	Bool	fError;
static	GHashTable	*Styles;

#define	GetSymbol	(StyleToken = StyleLex(FALSE))
#define	GetName		(StyleToken = StyleLex(TRUE))

static	void
_Error(
	char	*msg,
	char	*fn,
	int		line)
{
	printf("%s:%d:%s\n",fn,line,msg);
}
#undef	Error
#define	Error(msg)	{fError=TRUE;_Error((msg),StyleFileName,StylecLine);}

static	int
HexStringToInt(
	char	*p,
	size_t	len)
{
	int		ret;

	ret = 0;
	while	(  len  >  0  ) {
		ret *= 16;
		if		(	(  *p  >=  '0'  )
				&&	(  *p  <=  '9'  ) ) {
			ret += *p - '0';
		} else
		if		(	(  *p  >=  'a'  )
				&&	(  *p  <=  'f'  ) ) {
			ret += *p - 'a' + 10;
		} else
		if		(	(  *p  >=  'A'  )
				&&	(  *p  <=  'F'  ) ) {
			ret += *p - 'A' + 10;
		}
		p ++;
		len --;
	}
	return	(ret);
}

static	void
ParColor(
	GdkColor	*color)
{
	size_t	len;
	char	*p;

	if		(  GetSymbol  ==  '{'  ) {
		if		(  GetSymbol  ==  T_ICONST  ) {
			color->red = StyleComInt;
		} else {
			Error("red invalid");
		}
		if		(  GetSymbol  !=  ','  ) {
			Error(", missing");
		}
		if		(  GetSymbol  ==  T_ICONST  ) {
			color->green = StyleComInt;
		} else {
			Error("red invalid");
		}
		if		(  GetSymbol  !=  ','  ) {
			Error(", missing");
		}
		if		(  GetSymbol  ==  T_ICONST  ) {
			color->blue = StyleComInt;
		} else {
			Error("red invalid");
		}
		if		(  GetSymbol  !=  '}'  ) {
			Error("} not found");
		}
	} else
	if		(  StyleToken  ==  T_SCONST  ) {
		p = StyleComSymbol;
		len = strlen(StyleComSymbol) / 3;
		color->red = HexStringToInt(p,len);
		p += len;
		color->green = HexStringToInt(p,len);
		p += len;
		color->blue = HexStringToInt(p,len);
		switch	(len) {
		  case	1:
			color->red *= 4369;
			color->green *= 4369;
			color->blue *= 4369;
			break;
		  case	2:
			color->red *= 257;
			color->green *= 257;
			color->blue *= 257;
			break;
		  case	3:
			color->red *= 16;
			color->green *= 16;
			color->blue *= 16;
			break;
		}
	}
}

static	void
ParStyle(void)
{
	int			token;
	GtkStyle	*st
	,			*dst;
	int			state;	

dbgmsg(">ParStyle");
	dst = gtk_widget_get_default_style();
	GetSymbol;
	do {
		switch	(StyleToken) {
		  case	T_STYLE:
			if		(  GetSymbol  ==  T_SCONST  ) {
				if		(  ( st = g_hash_table_lookup(Styles,StyleComSymbol) )
						   ==  NULL  ) {
					st = gtk_style_copy(dst);
					g_hash_table_insert(Styles,StrDup(StyleComSymbol),st);
				}
			} else {
				Error("style name not found");
				st = gtk_style_copy(dst);	/*	dummy	*/
			}
			if		(  GetSymbol  !=  '{'  ) {
				Error("{ missing");
			}
			GetSymbol;
			do {
				switch	(StyleToken) {
				  case	T_FONTSET:
					if		(  GetSymbol  !=  '='  ) {
						Error("= not found");
					}
					if		(  GetSymbol  ==  T_SCONST  ) {
						st->font = gdk_fontset_load(StyleComSymbol);
					} else {
						Error("font set not found");
					}
					break;
				  case	T_FG:
				  case	T_BG:
				  case	T_BG_PIXMAP:
					token = StyleToken;
					state = 0;
					if		(  GetSymbol  ==  '['  ) {
						switch	(GetSymbol) {
						  case	T_NORMAL:
							state = GTK_STATE_NORMAL;
							break;
						  case	T_ACTIVE:
							state = GTK_STATE_ACTIVE;
							break;
						  case	T_INSENSITIVE:
							state = GTK_STATE_INSENSITIVE;
							break;
						  case	T_PRELIGHT:
							state = GTK_STATE_PRELIGHT;
							break;
						  case	T_SELECTED:
							state = GTK_STATE_SELECTED;
							break;
						  default:
							Error("state missing");
							break;
						}
						if		(  GetSymbol  !=  ']'  ) {
							Error("] missing");
						}
					} else {
						Error("[ missing");
					}
					if		(  GetSymbol  !=  '='  ) {
						Error("= not found");
					}
					switch	(token) {
					  case	T_FG:
						ParColor(&st->fg[state]);
						break;
					  case	T_BG:
						ParColor(&st->bg[state]);
						break;
					  case	T_BG_PIXMAP:
						Error("not support");
						break;
					}
					break;
				}
				GetSymbol;
			}	while	(  StyleToken  !=  '}'  );
			break;
		  default:
			Error("style missing");
			printf("token = %d(%c)\n",StyleToken,StyleToken);
			break;
		}
		GetSymbol;
	}	while	(  StyleToken  !=  T_EOF  );
dbgmsg("<ParStyle");
}

extern	void
StyleParserInit(void)
{
	StyleLexInit();
	Styles = NewNameHash();	 
}

extern	void
StyleParser(
	char	*name)
{
	FILE	*fp;
	struct	stat	stbuf;

dbgmsg(">StyleParser");
	if		(  stat(name,&stbuf)  ==  0  ) { 
		fError = FALSE;
		StylecLine = 1;
		if		(  ( fp = fopen(name,"r") )  !=  NULL  ) {
			StyleFileName = name;
			StyleFile = fp;
			ParStyle();
			fclose(fp);
		}
	}
dbgmsg("<StyleParser");
}

extern	GtkStyle	*
GetStyle(
	char	*name)
{
	GtkStyle	*p;

	if		(  ( p = g_hash_table_lookup(Styles,name) )  ==  NULL  ) {
		p = gtk_widget_get_default_style();
	}
	return	(p);
}

#ifdef	MAIN
static	char	*state[] = {
	"NORMAL", "ACTIVE", "PRELIGHT", "SELECTED", "INSENSITIVE" };

static	void
PrintStyle(
	GtkStyle	*value)
{
	int		i;

	printf("fontset = [%X]\n",(int)value->font);
	for	( i = 0 ; i < 5 ; i ++ ) {
		printf("fg[%s] = { %d, %d, %d }\n",state[i],
			   value->fg[i].red,value->fg[i].green,value->fg[i].blue);
		printf("bg[%s] = { %d, %d, %d }\n",state[i],
			   value->bg[i].red,value->bg[i].green,value->bg[i].blue);
	}
}
static	void
Dump(
	char		*name,
	GtkStyle	*value,
	gpointer	user_data)
{
	printf("name = [%s]\n",name);
	PrintStyle(value);
}
	
extern	int
main(
	int		argc,
	char	**argv)
{
	gtk_set_locale();
	gtk_init(&argc, &argv);
	
	StyleParserInit();
	StyleParser("testrc");

	printf("*** dump ***\n");
	g_hash_table_foreach(Styles,(GHFunc)Dump,NULL);
	printf("*** dump ***\n");
	PrintStyle(GetStyle("green"));

	return	(0);
}
#endif
